from typing import TYPE_CHECKING

import numpy as np
import torch

from eir.models.model_training_utils import predict_on_batch
from eir.serve_modules.serve_experiment_io import ServeExperiment
from eir.setup.schema_modules.output_schemas_array import ArrayOutputTypeConfig
from eir.setup.schema_modules.output_schemas_image import ImageOutputTypeConfig
from eir.setup.schema_modules.output_schemas_sequence import (
    SequenceOutputSamplingConfig,
)
from eir.setup.schemas import SequenceOutputTypeConfig
from eir.train_utils.evaluation_handlers.train_handlers_array_output import (
    ArrayOutputEvalSample,
    one_shot_array_generation,
    reverse_diffusion_array_generation,
)
from eir.train_utils.evaluation_handlers.train_handlers_sequence_output import (
    SequenceOutputEvalSample,
    autoregressive_sequence_generation,
)

if TYPE_CHECKING:
    from eir.serve_modules.serve_input_setup import ServeBatch

al_merged_predictions = list[
    dict[str, dict[str, torch.Tensor | list[int] | np.ndarray]]
]


@torch.inference_mode()
def run_serve_prediction(
    serve_experiment: ServeExperiment,
    batch: "ServeBatch",
) -> list[dict[str, dict[str, torch.Tensor | list[int] | np.ndarray]]]:
    one_shot_prediction = predict_on_batch(
        model=serve_experiment.model,
        inputs=batch.inputs,
    )

    sequence_predictions = _run_serve_sequence_generation(
        serve_experiment=serve_experiment,
        batch=batch,
    )

    array_predictions = _run_serve_array_generation(
        serve_experiment=serve_experiment,
        batch=batch,
    )

    predictions = merge_predictions_per_sample(
        one_shot_prediction=one_shot_prediction,
        sequence_predictions=sequence_predictions,
        array_predictions=array_predictions,
        batch_size=len(batch.ids),
    )

    return predictions


def merge_predictions_per_sample(
    one_shot_prediction: dict[str, dict[str, torch.Tensor]],
    sequence_predictions: dict[str, dict[str, list[list[int]]]],
    array_predictions: dict[str, dict[str, list[np.ndarray]]],
    batch_size: int,
) -> al_merged_predictions:

    merged_predictions: al_merged_predictions = [{} for _ in range(batch_size)]

    for os_key, os_value in one_shot_prediction.items():
        for sub_key, tensor in os_value.items():
            for i in range(batch_size):
                if os_key not in merged_predictions[i]:
                    merged_predictions[i][os_key] = {}
                merged_predictions[i][os_key][sub_key] = tensor[i]

    for seq_key, seq_value in sequence_predictions.items():
        for sub_key, seq_list in seq_value.items():
            for i in range(batch_size):
                if seq_key not in merged_predictions[i]:
                    merged_predictions[i][seq_key] = {}
                merged_predictions[i][seq_key][sub_key] = seq_list[i]

    for arr_key, arr_value in array_predictions.items():
        for sub_key, array_list in arr_value.items():
            for i in range(batch_size):
                if arr_key not in merged_predictions[i]:
                    merged_predictions[i][arr_key] = {}
                merged_predictions[i][arr_key][sub_key] = array_list[i]

    return merged_predictions


def _run_serve_sequence_generation(
    serve_experiment: ServeExperiment,
    batch: "ServeBatch",
) -> dict[str, dict[str, list[list[int]]]]:
    """
    Note that we always expect two levels for the predictions, in the case
    of tabular that's the output name and the column name, in the case of
    sequence we simply output name twice to conform to the same structure.
    """
    prepared = {}

    output_configs = serve_experiment.configs.output_configs
    for config in output_configs:
        if config.output_info.output_type != "sequence":
            continue

        cur_output_name = config.output_info.output_name

        output_type_info = config.output_type_info
        assert isinstance(output_type_info, SequenceOutputTypeConfig)

        assert config.sampling_config is not None

        if output_type_info.sequence_operation == "autoregressive":
            assert isinstance(config.sampling_config, SequenceOutputSamplingConfig)

            eval_samples = _build_sequence_eval_samples_batch(
                serve_batch=batch,
            )

            hooks = serve_experiment.hooks
            assert hooks is not None

            generated_tokens = autoregressive_sequence_generation(
                input_objects=serve_experiment.inputs,
                eval_samples=eval_samples,
                seq_output_name=cur_output_name,
                experiment=serve_experiment,
                default_eir_hooks=hooks,
                sampling_config=config.sampling_config,
            )

            prepared[cur_output_name] = {cur_output_name: generated_tokens}

    return prepared


def _build_sequence_eval_samples_batch(
    serve_batch: "ServeBatch",
) -> tuple[SequenceOutputEvalSample, ...]:
    eval_samples = []

    for i in range(len(serve_batch.ids)):

        prepared_inputs = serve_batch.inputs_split[i]

        eval_sample = SequenceOutputEvalSample(
            inputs_to_model=prepared_inputs,
            target_labels={},
            sample_id=serve_batch.ids[i],
        )
        eval_samples.append(eval_sample)

    return tuple(eval_samples)


def _run_serve_array_generation(
    serve_experiment: ServeExperiment,
    batch: "ServeBatch",
) -> dict[str, dict[str, list[np.ndarray]]]:
    prepared = {}

    output_configs = serve_experiment.configs.output_configs

    for config in output_configs:
        if config.output_info.output_type not in ("array", "image"):
            continue

        cur_output_name = config.output_info.output_name
        output_type_info = config.output_type_info
        assert isinstance(
            output_type_info, (ArrayOutputTypeConfig, ImageOutputTypeConfig)
        )

        assert config.sampling_config is not None

        eval_samples = _build_array_eval_samples_batch(serve_batch=batch)

        hooks = serve_experiment.hooks
        assert hooks is not None

        is_diffusion = output_type_info.loss == "diffusion"

        if is_diffusion:
            time_steps = output_type_info.diffusion_time_steps
            assert time_steps is not None
            array_outputs = reverse_diffusion_array_generation(
                eval_samples=eval_samples,
                array_output_name=cur_output_name,
                experiment=serve_experiment,
                default_eir_hooks=hooks,
                num_steps=time_steps,
            )
        else:
            array_outputs = one_shot_array_generation(
                eval_samples=eval_samples,
                array_output_name=cur_output_name,
                experiment=serve_experiment,
                default_eir_hooks=hooks,
            )

        prepared[cur_output_name] = {cur_output_name: array_outputs}

    return prepared


def _build_array_eval_samples_batch(
    serve_batch: "ServeBatch",
) -> tuple[ArrayOutputEvalSample, ...]:
    eval_samples = []

    for i in range(len(serve_batch.ids)):
        eval_sample = ArrayOutputEvalSample(
            inputs_to_model=serve_batch.inputs_split[i],
            target_labels={},
            sample_id=serve_batch.ids[i],
        )
        eval_samples.append(eval_sample)

    return tuple(eval_samples)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np

from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.input_setup_modules.setup_array import (
    ArrayNormalizationStats,
    get_array_normalization_values,
    get_data_dimension_from_data_source,
    get_dtype_from_data_source,
)
from eir.setup.schemas import ArrayOutputTypeConfig, OutputConfig
from eir.train_utils.step_modules.diffusion import (
    DiffusionConfig,
    initialize_diffusion_config,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class ComputedArrayOutputInfo:
    output_config: OutputConfig
    data_dimensions: DataDimensions
    dtype: np.dtype
    normalization_stats: Optional[ArrayNormalizationStats] = None
    diffusion_config: Optional[DiffusionConfig] = None


def set_up_array_output(
    output_config: OutputConfig, *args, **kwargs
) -> ComputedArrayOutputInfo:
    data_dimensions = get_data_dimension_from_data_source(
        data_source=Path(output_config.output_info.output_source),
        deeplake_inner_key=output_config.output_info.output_inner_key,
    )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, ArrayOutputTypeConfig)

    normalization_stats: Optional[ArrayNormalizationStats] = None
    if output_type_info.normalization is not None:
        normalization_stats = get_array_normalization_values(
            source=output_config.output_info.output_source,
            inner_key=output_config.output_info.output_inner_key,
            normalization=output_type_info.normalization,
            data_dimensions=data_dimensions,
            max_samples=output_type_info.adaptive_normalization_max_samples,
        )

    dtype = get_dtype_from_data_source(
        data_source=Path(output_config.output_info.output_source),
        deeplake_inner_key=output_config.output_info.output_inner_key,
    )

    diffusion_config = None
    if output_type_info.loss == "diffusion":
        time_steps = output_type_info.diffusion_time_steps
        if time_steps is None:
            raise ValueError(
                "Diffusion loss requires specifying the number of time steps."
                "Please set `diffusion_time_steps` in the output config."
            )
        diffusion_config = initialize_diffusion_config(time_steps=time_steps)

    array_output_object = ComputedArrayOutputInfo(
        output_config=output_config,
        data_dimensions=data_dimensions,
        normalization_stats=normalization_stats,
        dtype=dtype,
        diffusion_config=diffusion_config,
    )

    return array_output_object



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Protocol,
    Sequence,
    Type,
    Union,
)

from eir.experiment_io.experiment_io import get_run_folder_from_model_path
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup import schemas
from eir.setup.input_setup_modules.setup_array import (
    ComputedArrayInputInfo,
    set_up_array_input,
)
from eir.setup.input_setup_modules.setup_bytes import (
    ComputedBytesInputInfo,
    set_up_bytes_input_for_training,
)
from eir.setup.input_setup_modules.setup_image import (
    ComputedImageInputInfo,
    set_up_image_input_for_training,
)
from eir.setup.input_setup_modules.setup_omics import (
    ComputedOmicsInputInfo,
    set_up_omics_input,
)
from eir.setup.input_setup_modules.setup_pretrained import (
    get_input_setup_from_pretrained_function_map,
)
from eir.setup.input_setup_modules.setup_sequence import (
    ComputedSequenceInputInfo,
    set_up_sequence_input_for_training,
)
from eir.setup.input_setup_modules.setup_tabular import (
    ComputedTabularInputInfo,
    set_up_tabular_input_for_training,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train_utils.step_logic import Hooks


logger = get_logger(__name__)

al_input_objects = Union[
    ComputedOmicsInputInfo,
    ComputedTabularInputInfo,
    ComputedSequenceInputInfo,
    ComputedBytesInputInfo,
    ComputedImageInputInfo,
    ComputedArrayInputInfo,
    ComputedPredictTabularInputInfo,
    ComputedServeTabularInputInfo,
]
al_input_objects_as_dict = Dict[str, al_input_objects]

al_serializable_input_objects = Union[
    ComputedSequenceInputInfo,
    ComputedImageInputInfo,
    ComputedBytesInputInfo,
    ComputedArrayInputInfo,
]

al_serializable_input_classes = Union[
    Type[ComputedSequenceInputInfo],
    Type[ComputedImageInputInfo],
    Type[ComputedBytesInputInfo],
    Type[ComputedArrayInputInfo],
]


class InputSetupFunction(Protocol):
    def __call__(
        self,
        input_config: schemas.InputConfig,
        hooks: Optional["Hooks"],
        **kwargs,
    ) -> al_input_objects: ...


class InputSetupGetterFunction(Protocol):
    def __call__(self, input_config: schemas.InputConfig) -> InputSetupFunction: ...


def set_up_inputs_general(
    inputs_configs: schemas.al_input_configs,
    hooks: Optional["Hooks"],
    setup_func_getter: InputSetupGetterFunction,
    setup_func_kwargs: Dict[str, Any],
) -> al_input_objects_as_dict:
    all_inputs = {}

    name_config_iter = get_input_name_config_iterator(input_configs=inputs_configs)

    for name, input_config in name_config_iter:
        setup_func = setup_func_getter(input_config=input_config)

        cur_input_data_config = input_config.input_info
        logger.info(
            "Setting up %s inputs '%s' from %s.",
            cur_input_data_config.input_type,
            cur_input_data_config.input_name,
            cur_input_data_config.input_source,
        )

        set_up_input = setup_func(
            input_config=input_config, hooks=hooks, **setup_func_kwargs
        )
        all_inputs[name] = set_up_input

    return all_inputs


def set_up_inputs_for_training(
    inputs_configs: schemas.al_input_configs,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    hooks: Union["Hooks", None],
) -> al_input_objects_as_dict:
    train_input_setup_kwargs = {
        "train_ids": train_ids,
        "valid_ids": valid_ids,
    }
    all_inputs = set_up_inputs_general(
        inputs_configs=inputs_configs,
        hooks=hooks,
        setup_func_getter=get_input_setup_function_for_train,
        setup_func_kwargs=train_input_setup_kwargs,
    )

    return all_inputs


def get_input_name_config_iterator(input_configs: schemas.al_input_configs):
    for input_config in input_configs:
        cur_input_data_config = input_config.input_info
        cur_name = cur_input_data_config.input_name
        yield cur_name, input_config


def get_input_setup_function_for_train(
    input_config: schemas.InputConfig,
) -> Callable[..., al_input_objects]:
    input_type = input_config.input_info.input_type
    pretrained_config = input_config.pretrained_config

    from_scratch_mapping = get_input_setup_function_map()

    if pretrained_config:
        pretrained_run_folder = get_run_folder_from_model_path(
            model_path=pretrained_config.model_path
        )
        from_pretrained_mapping = get_input_setup_from_pretrained_function_map(
            run_folder=pretrained_run_folder,
            load_module_name=pretrained_config.load_module_name,
        )
        return from_pretrained_mapping[input_type]

    return from_scratch_mapping[input_type]


def get_input_setup_function_map() -> dict[str, Callable[..., al_input_objects]]:
    setup_mapping: dict[str, Callable[..., al_input_objects]] = {
        "omics": set_up_omics_input,
        "tabular": set_up_tabular_input_for_training,
        "sequence": set_up_sequence_input_for_training,
        "bytes": set_up_bytes_input_for_training,
        "image": set_up_image_input_for_training,
        "array": set_up_array_input,
    }

    return setup_mapping



from pathlib import Path
from typing import Optional, Union

import numpy as np
import torch
from PIL.Image import Image, fromarray

from eir.data_load.data_preparation_modules.common import _load_deeplake_sample
from eir.data_load.data_source_modules import deeplake_ops
from eir.setup.input_setup_modules.setup_image import (
    ComputedImageInputInfo,
    default_image_loader,
)
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo


def prepare_image_data(
    image_input_object: Union["ComputedImageInputInfo", "ComputedImageOutputInfo"],
    image_data: Image,
    test_mode: bool,
) -> torch.Tensor:
    """
    The transforms take care of converting the image object to a copied tensor.
    """

    image_data_clone = image_data.copy()

    if test_mode:
        image_prepared = image_input_object.base_transforms(img=image_data_clone)
    else:
        image_prepared = image_input_object.all_transforms(img=image_data_clone)

    return image_prepared


def image_load_wrapper(
    data_pointer: Union[Path, int],
    input_source: str,
    image_mode: Optional[str],
    deeplake_inner_key: Optional[str] = None,
) -> Image:
    """
    Squeeze there since loading a saved 2D image, e.g. saved as (16, 16),
    in deeplake will be loaded as (16, 16, 1).
    """
    if deeplake_ops.is_deeplake_dataset(data_source=input_source):
        assert deeplake_inner_key is not None
        assert isinstance(data_pointer, int)
        image_data = _load_deeplake_sample(
            data_pointer=data_pointer,
            input_source=input_source,
            inner_key=deeplake_inner_key,
        )
        array = np.uint8(image_data * 255).squeeze(axis=-1)  # type: ignore
        pil_image = fromarray(obj=array)
    else:
        assert isinstance(data_pointer, Path)
        pil_image = default_image_loader(path=str(data_pointer))

    if image_mode is not None:
        pil_image = pil_image.convert(mode=image_mode)

    return pil_image



import atexit
import json
import shutil
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Sequence, Union
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import websocket
from aislib.misc_utils import ensure_path_exists
from PIL import Image
from tqdm import tqdm

from eir.serve_modules.serve_network_utils import deserialize_array, deserialize_image
from eir.setup.config import Configs
from eir.setup.schemas import InputConfig, OutputConfig
from eir.setup.streaming_data_setup.protocol import PROTOCOL_VERSION
from eir.setup.streaming_data_setup.streaming_data_utils import (
    connect_to_server,
    receive_with_timeout,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def cleanup_streaming_setup(path: Path) -> None:
    if path.exists():
        logger.info(f"Cleaning up streaming setup folder: {path}")
        try:
            shutil.rmtree(path)
            logger.info("Cleanup completed successfully.")
        except Exception as e:
            logger.error(f"Error during cleanup: {e}")
    else:
        logger.info(f"No cleanup needed. Folder does not exist: {path}")


class StreamDataGatherer:
    def __init__(
        self,
        websocket_url: str,
        output_folder: str,
        input_configs: Dict[str, Any],
        output_configs: Dict[str, Any],
        batch_size: int = 32,
        max_samples: int = 1_000,
    ):
        self.websocket_url = websocket_url
        self.output_folder = output_folder
        self.inputs = input_configs
        self.outputs = output_configs
        self.batch_size = batch_size
        self.max_samples = max_samples
        self.base_path = Path(f"{output_folder}/streaming_setup")
        self.dataset_info = None
        self.input_dataframes: dict[str, pd.DataFrame] = {}
        self.output_dataframes: dict[str, pd.DataFrame] = {}

        atexit.register(cleanup_streaming_setup, self.base_path)

    def get_dataset_info(self, ws: websocket.WebSocket):
        ws.send(json.dumps({"type": "getInfo"}))

        info_data = receive_with_timeout(websocket=ws)

        if info_data["type"] != "info":
            raise ValueError(f"Unexpected response type: {info_data['type']}")

        self.dataset_info = info_data["payload"]
        logger.info("Received dataset information.")

    def gather_and_save_data(self):
        self.base_path.mkdir(parents=True, exist_ok=True)

        with connect_to_server(
            websocket_url=self.websocket_url,
            protocol_version=PROTOCOL_VERSION,
        ) as ws:
            self.get_dataset_info(ws=ws)

            total_samples = 0
            pbar = tqdm(
                total=self.max_samples,
                desc="Gathering samples information and statistics",
                unit=" sample",
            )
            while total_samples < self.max_samples:
                ws.send(
                    json.dumps(
                        {
                            "type": "getData",
                            "payload": {"batch_size": self.batch_size},
                        }
                    )
                )

                batch_data = receive_with_timeout(websocket=ws)

                if batch_data["type"] != "data":
                    logger.error(f"Unexpected response type: {batch_data['type']}")
                    continue

                samples = batch_data["payload"]

                assert self.dataset_info is not None
                for sample in samples:
                    processed_input = process_inputs(
                        sample_input=sample["inputs"],
                        dataset_info=self.dataset_info,
                    )
                    processed_output = process_outputs(
                        sample_output=sample["target_labels"],
                        dataset_info=self.dataset_info,
                    )

                    save_inputs(
                        processed_input=processed_input,
                        sample_id=sample["sample_id"],
                        base_path=self.base_path,
                        inputs=self.dataset_info["inputs"],
                        dataframes=self.input_dataframes,
                    )
                    save_outputs(
                        processed_output=processed_output,
                        sample_id=sample["sample_id"],
                        base_path=self.base_path,
                        outputs=self.dataset_info["outputs"],
                        dataframes=self.output_dataframes,
                    )

                    total_samples += 1
                    pbar.update(1)
                    if total_samples >= self.max_samples:
                        break

            pbar.close()
            logger.info(f"Finished processing {total_samples} samples.")

        for name, df in self.input_dataframes.items():
            output_path = self.base_path / "input" / name
            ensure_path_exists(path=output_path, is_folder=True)
            df.to_csv(output_path / f"{name}.csv", index=False)

        for name, df in self.output_dataframes.items():
            output_path = self.base_path / "output" / name
            ensure_path_exists(path=output_path, is_folder=True)
            df.to_csv(output_path / f"{name}.csv", index=False)

    def reset(self):
        with connect_to_server(
            websocket_url=self.websocket_url,
            protocol_version=PROTOCOL_VERSION,
        ) as ws:
            ws.send(
                json.dumps(
                    {
                        "type": "reset",
                        "payload": {},
                    },
                )
            )

            for _ in range(2):
                reset_message = receive_with_timeout(websocket=ws)

                if reset_message["type"] == "reset":
                    logger.info(
                        f"Received broadcast: {reset_message['payload']['message']}"
                    )
                elif reset_message["type"] == "resetConfirmation":
                    logger.info("Reset command sent successfully.")
                else:
                    logger.error(
                        f"Unexpected response to reset command: {reset_message}"
                    )

    def get_status(self):
        with connect_to_server(
            websocket_url=self.websocket_url,
            protocol_version=PROTOCOL_VERSION,
        ) as ws:
            ws.send(
                json.dumps(
                    {
                        "type": "status",
                        "payload": {},
                    },
                )
            )

            status_data = receive_with_timeout(websocket=ws)

            if status_data["type"] == "status":
                logger.info(f"Current status: {status_data['payload']}")
                return status_data["payload"]
            else:
                logger.error(f"Unexpected response to status request: {status_data}")


def process_single_data(data: Any, info: Dict[str, Any]) -> Any:
    data_type = info["type"]

    if data_type in ["array", "omics"]:
        dtype = np.dtype(bool) if data_type == "omics" else np.dtype(np.float32)
        return process_array(data=data, shape=info["shape"], dtype=dtype)
    elif data_type == "image":
        return process_image(data=data, mode=info.get("mode"))
    elif data_type in ["tabular", "sequence"]:
        return data
    else:
        logger.warning(f"Unsupported data type: {data_type}")
        return data


def process_array(data: str, shape: list[int], dtype: np.dtype) -> np.ndarray:
    array = deserialize_array(
        array_str=data,
        dtype=dtype,
        shape=tuple(shape),
    )

    return array


def process_image(
    data: str, mode: Optional[Literal["L", "RGB", "RGBA"]]
) -> Image.Image:
    image = deserialize_image(
        image_str=data,
        image_mode=mode,
    )

    return image


def process_inputs(
    sample_input: dict[str, Any],
    dataset_info: dict[str, Any],
) -> dict[str, Any]:
    processed_inputs = {}
    for input_name, input_data in sample_input.items():
        input_info = dataset_info["inputs"].get(input_name)

        if input_info:
            processed_inputs[input_name] = process_single_data(
                data=input_data,
                info=input_info,
            )
        else:
            logger.warning(f"No info found for input {input_name}")
            processed_inputs[input_name] = input_data
    return processed_inputs


def process_outputs(
    sample_output: Dict[str, Any], dataset_info: Dict[str, Any]
) -> Dict[str, Any]:
    processed_outputs = {}
    for output_name, output_data in sample_output.items():
        output_info = dataset_info["outputs"].get(output_name)
        output_type = output_info["type"]

        if output_type in ["array", "sequence", "image"]:
            output_data = output_data[output_name]

        if output_info:
            processed_outputs[output_name] = process_single_data(
                data=output_data,
                info=output_info,
            )
        else:
            logger.warning(f"No info found for output {output_name}")
            processed_outputs[output_name] = output_data
    return processed_outputs


def save_data(
    data_type: str,
    data_name: str,
    data: Any,
    save_path: Path,
    sample_id: str,
    dataframes: Dict[str, pd.DataFrame],
):
    if data_type in ["omics", "array"]:
        assert isinstance(data, np.ndarray)
        np.save(save_path / f"{sample_id}.npy", data)
    elif data_type == "image":
        assert isinstance(data, Image.Image)
        data.save(save_path / f"{sample_id}.png")
    elif data_type in ["tabular", "sequence"]:
        if data_name not in dataframes:
            dataframes[data_name] = pd.DataFrame(columns=["ID"])

        df = dataframes[data_name]
        if data_type == "tabular":
            assert isinstance(data, dict)
            new_row = pd.DataFrame({"ID": [sample_id], **data})
        else:
            assert isinstance(data, str)
            new_row = pd.DataFrame({"ID": [sample_id], "Sequence": [data]})

        dataframes[data_name] = pd.concat([df, new_row], ignore_index=True)
    else:
        logger.warning(f"Unsupported data type: {data_type}")


def save_inputs(
    processed_input: Dict[str, Any],
    sample_id: str,
    base_path: Path,
    inputs: Dict[str, Dict],
    dataframes: Dict[str, pd.DataFrame],
):
    for input_name, input_config in inputs.items():
        input_type = input_config["type"]
        input_data = processed_input.get(input_name)

        if input_data is None:
            logger.warning(f"No data for input {input_name} in sample {sample_id}")
            continue

        save_path = base_path / "input" / input_name
        save_path.mkdir(parents=True, exist_ok=True)

        save_data(
            data_type=input_type,
            data_name=input_name,
            data=input_data,
            save_path=save_path,
            sample_id=sample_id,
            dataframes=dataframes,
        )


def save_outputs(
    processed_output: Dict[str, Any],
    sample_id: str,
    base_path: Path,
    outputs: Dict[str, Dict],
    dataframes: Dict[str, pd.DataFrame],
):
    for output_name, output_config in outputs.items():
        output_type = output_config["type"]
        output_data = processed_output.get(output_name)

        if output_data is None:
            logger.warning(f"No data for output {output_name} in sample {sample_id}")
            continue

        save_path = base_path / "output" / output_name
        save_path.mkdir(parents=True, exist_ok=True)

        save_data(
            data_type=output_type,
            data_name=output_name,
            data=output_data,
            save_path=save_path,
            sample_id=sample_id,
            dataframes=dataframes,
        )


def validate_streaming_setup(configs: Configs) -> Optional[str]:
    def is_websocket_url(url: str) -> bool:
        parsed = urlparse(url)
        return parsed.scheme in ("ws", "wss")

    websocket_address = None
    sources_to_check = []

    for input_config in configs.input_configs:
        sources_to_check.append(input_config.input_info.input_source)

    for output_config in configs.output_configs:
        sources_to_check.append(output_config.output_info.output_source)

    for source in sources_to_check:
        if is_websocket_url(source):
            if websocket_address is None:
                websocket_address = source
            elif source != websocket_address:
                raise ValueError(
                    f"Inconsistent WebSocket addresses found. "
                    f"Expected {websocket_address}, but found {source}"
                )
        elif websocket_address is not None:
            raise ValueError(
                f"Mixed source types found. "
                f"Expected all sources to be WebSocket if one is present. "
                f"Found WebSocket {websocket_address} and non-WebSocket {source}"
            )

    return websocket_address


def gather_streaming_data_for_setup(
    websocket_url: str,
    output_folder: str,
    configs: Configs,
    batch_size: int,
    max_samples: int,
) -> tuple[str, Path]:
    input_configs = configs.input_configs
    output_configs = configs.output_configs

    inputs = {}
    for input_config in input_configs:
        input_name = input_config.input_info.input_name
        input_type = input_config.input_info.input_type
        inputs[input_name] = {"type": input_type}

    outputs = {}
    for output_config in output_configs:
        output_name = output_config.output_info.output_name
        output_type = output_config.output_info.output_type
        outputs[output_name] = {"type": output_type}

    gatherer = StreamDataGatherer(
        websocket_url=websocket_url,
        output_folder=output_folder,
        input_configs=inputs,
        output_configs=outputs,
        batch_size=batch_size,
        max_samples=max_samples,
    )

    gatherer.get_status()

    gatherer.gather_and_save_data()

    gatherer.reset()

    gatherer.get_status()

    return websocket_url, gatherer.base_path


def patch_configs_for_local_data(
    configs: Configs,
    local_data_path: Path,
) -> tuple[Configs, Configs]:
    patched_configs = deepcopy(configs)

    def update_source(
        config: Union[InputConfig, OutputConfig],
        config_type: str,
    ) -> None:
        match config:
            case InputConfig():
                name = config.input_info.input_name
                data_type = config.input_info.input_type
            case OutputConfig():
                name = config.output_info.output_name
                data_type = config.output_info.output_type
            case _:
                raise ValueError("Unsupported config type")

        new_source: str | Path
        if data_type in ["tabular", "sequence"]:
            new_source = local_data_path / config_type / name / f"{name}.csv"
        elif data_type in ["omics", "array", "image"]:
            new_source = str(local_data_path / config_type / name)
        else:
            raise ValueError(f"Unsupported data type: {data_type}")

        if config_type == "input":
            assert isinstance(config, InputConfig)
            config.input_info.input_source = str(new_source)
        else:
            assert isinstance(config, OutputConfig)
            config.output_info.output_source = str(new_source)

    for input_config in patched_configs.input_configs:
        update_source(config=input_config, config_type="input")

    for output_config in patched_configs.output_configs:
        update_source(config=output_config, config_type="output")

    patched_configs.input_configs = _inject_correct_sequence_input_from_linked_output(
        input_configs=patched_configs.input_configs,
        output_configs=patched_configs.output_configs,
    )

    patched_configs.global_config.basic_experiment.n_epochs = 1

    return configs, patched_configs


def _inject_correct_sequence_input_from_linked_output(
    input_configs: Sequence[InputConfig],
    output_configs: Sequence[OutputConfig],
) -> Sequence[InputConfig]:
    """
    This is needed as EIR earlier creates a matching input sequence configuration
    for sequence outputs. While the source is correctly set at that point, we
    override that when patching the local data paths, so we need to re-inject the
    correct source for the input sequence
    """
    new_input_configs = deepcopy(input_configs)

    for output_config in output_configs:
        if output_config.output_info.output_type == "sequence":
            output_name = output_config.output_info.output_name
            matching_input = next(
                (
                    input_config
                    for input_config in new_input_configs
                    if input_config.input_info.input_name == output_name
                ),
                None,
            )
            assert (
                matching_input is not None
            ), f"No matching input found for {output_name}"

            matching_input.input_info.input_source = (
                output_config.output_info.output_source
            )

    return new_input_configs


def setup_and_gather_streaming_data(
    output_folder: str,
    configs: Configs,
    batch_size: int,
    max_samples: int,
) -> Optional[tuple[str, Path]]:
    try:
        websocket_url = validate_streaming_setup(configs=configs)
        if websocket_url:
            return gather_streaming_data_for_setup(
                websocket_url=websocket_url,
                output_folder=output_folder,
                configs=configs,
                batch_size=batch_size,
                max_samples=max_samples,
            )
        else:
            logger.info(
                "No streaming setup detected. Proceeding with standard data loading."
            )
            return None
    except ValueError as e:
        logger.error(f"Invalid streaming setup: {str(e)}")
        return None
